#include <stdlib.h>
#include <gl/gl.h>
#include <gl/device.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include "forms.h"
#include "movforms.h"

/****
   This is a simple application that tests the speed of the
   computer.  It draws some rotating objects, using
   different materials and different transformations.

   It uses the forms library to do the interaction with the user.

   Written by Mark Overmars
   Date  5 May 1991
****/

/******************************* THE OBJECTS ********************************/

typedef struct {
   float c[3];
} POINT;

typedef struct {
    int numb;   /* number of vertices */
    POINT v[4]; /* vertices */
    POINT n[4]; /* normals */
} PATCH;

typedef struct {
    int numb;     /* number of patches */
    PATCH *pat;   /* the patches */
} OBJECT;

OBJECT sphere,cube,cylinder,cone,glass,pyramid;

/*************************
 Making Spin Objects
*************************/

#define PI    3.14159265
#define ZERO  0.00001

void makespinobject(OBJECT *obj, int smooth, int rot, int n,
		float x1[], float z1[], float nx1[], float nz1[],
		float x2[], float z2[], float nx2[], float nz2[])
/* makes a spin object */
{
    float th,dth,a1,a2;
    PATCH *p;
    int i,j;
    obj->numb = rot*n;
    p = obj->pat = (PATCH *) malloc(obj->numb * sizeof(PATCH));
    dth = 2.0*PI/rot;
    for (i=0; i<n; i++)
    {
        for (j=0; j<rot; j++)
        {
 	    th =  dth * j;   
	    if (smooth) {a1 = th; a2 =th+dth;} else { a1 = a2 = th +dth/2.0;};
	    if (x1[i] < ZERO || x2[i] <ZERO) {p->numb=3;} else {p->numb = 4;};
	    p->v[0].c[0] = x1[i]*sin(th);     p->n[0].c[0] = nx1[i]*sin(a1);
	    p->v[0].c[1] = x1[i]*cos(th);     p->n[0].c[1] = nx1[i]*cos(a1);
	    p->v[0].c[2] = z1[i];             p->n[0].c[2] = nz1[i];
	    p->v[1].c[0] = x1[i]*sin(th+dth); p->n[1].c[0] = nx1[i]*sin(a2);
	    p->v[1].c[1] = x1[i]*cos(th+dth); p->n[1].c[1] = nx1[i]*cos(a2);
	    p->v[1].c[2] = z1[i];             p->n[1].c[2] = nz1[i];
	    p->v[2].c[0] = x2[i]*sin(th+dth); p->n[2].c[0] = nx2[i]*sin(a2);
	    p->v[2].c[1] = x2[i]*cos(th+dth); p->n[2].c[1] = nx2[i]*cos(a2);
	    p->v[2].c[2] = z2[i];             p->n[2].c[2] = nz2[i];
	    p->v[3].c[0] = x2[i]*sin(th);     p->n[3].c[0] = nx2[i]*sin(a1);
	    p->v[3].c[1] = x2[i]*cos(th);     p->n[3].c[1] = nx2[i]*cos(a1);
	    p->v[3].c[2] = z2[i];             p->n[3].c[2] = nz2[i];
	    if (x1[i] <ZERO) 
	    {
		p->v[1] = p->v[2]; p->n[1] = p->n[2];
		p->v[2] = p->v[3]; p->n[2] = p->n[3];
	    };
	    p++;
	}
    }
}

void makesmoothspinobject(OBJECT *obj, int rot, int n, float x[], float z[])
/* makes a smooth, connected spin object */
{
    float x1[20],z1[20],nx1[20],nz1[20],x2[20],z2[20],nx2[20],nz2[20],l[20];
    int i;
    /* fill in the points */
    x1[0] = x[0]; z1[0] = z[0];
    for (i=1; i<n ; i++)
    	{ x1[i] = x2[i-1] = x[i]; z1[i] = z2[i-1] = z[i]; };
    /* fill in the normals */
    for (i=0; i<n-1 ; i++)
      {l[i] = sqrt((x[i+1]-x[i])*(x[i+1]-x[i]) + (z[i+1]-z[i])*(z[i+1]-z[i]));};
    nx1[0]   = 0.0; nz1[0]   =  1.0;
    nx2[n-2] = 0.0; nz2[n-2] = -1.0;
    for (i=1; i<n-1 ; i++)
    {
	nx1[i] = nx2[i-1] = ((z[i-1]-z[i])/l[i-1] + (z[i] - z[i+1])/l[i] )/2.0;
	nz1[i] = nz2[i-1] = ((x[i]-x[i-1])/l[i-1] + (x[i+1] - x[i])/l[i] )/2.0;
    }
    /* make it */
    makespinobject(obj,1,rot,n-1,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}

/*Making some nice objects */

float x1[20],z1[20],nx1[20],nz1[20],x2[20],z2[20],nx2[20],nz2[20];

void makesphere(OBJECT *sphere,int n)
{
    int i;
    x1[0] = nx1[0] = 0.0; z1[0] = nz1[0] = 1.0;
    for (i=0; i<n-1; i++)
    {
	x2[i] = nx2[i] = x1[i+1] = nx1[i+1] = sin((PI/n)*(1.0+i));
	z2[i] = nz2[i] = z1[i+1] = nz1[i+1] = cos((PI/n)*(1.0+i));
    }
    x2[n-1] = nx2[n-1] = 0.0; z2[n-1] = nz2[n-1] = -1.0;
    makespinobject(sphere,1,2*n,n,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}

void makecylinder(OBJECT *cylinder, int n)
{
    x1[0] =  0.0; nx1[0] = 0.0; z1[0] =  1.0; nz1[0] =  1.0;
    x2[0] =  1.0; nx2[0] = 0.0; z2[0] =  1.0; nz2[0] =  1.0;
    x1[1] =  1.0; nx1[1] = 1.0; z1[1] =  1.0; nz1[1] =  0.0;
    x2[1] =  1.0; nx2[1] = 1.0; z2[1] = -1.0; nz2[1] =  0.0;
    x1[2] =  1.0; nx1[2] = 0.0; z1[2] = -1.0; nz1[2] = -1.0;
    x2[2] =  0.0; nx2[2] = 0.0; z2[2] = -1.0; nz2[2] = -1.0;
    makespinobject(cylinder,1,2*n,3,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}

void makecone(OBJECT *cone, int n)
{
    x1[0] =  0.0; nx1[0] = 2.0/sqrt(5.0); z1[0] =  1.0; nz1[0] = 1.0/sqrt(5.0);
    x2[0] =  1.0; nx2[0] = 2.0/sqrt(5.0); z2[0] = -1.0; nz2[0] = 1.0/sqrt(5.0);
    x1[1] =  1.0; nx1[1] = 0.0;           z1[1] = -1.0; nz1[1] = -1.0;
    x2[1] =  0.0; nx2[1] = 0.0;           z2[1] = -1.0; nz2[1] = -1.0;
    makespinobject(cone,1,2*n,2,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}

void makecube(OBJECT *cube)
{
    x1[0] =  0.0;       nx1[0] = 0.0; z1[0] =  1.0; nz1[0] =  1.0;
    x2[0] =  sqrt(2.0); nx2[0] = 0.0; z2[0] =  1.0; nz2[0] =  1.0;
    x1[1] =  sqrt(2.0); nx1[1] = 1.0; z1[1] =  1.0; nz1[1] =  0.0;
    x2[1] =  sqrt(2.0); nx2[1] = 1.0; z2[1] = -1.0; nz2[1] =  0.0;
    x1[2] =  sqrt(2.0); nx1[2] = 0.0; z1[2] = -1.0; nz1[2] = -1.0;
    x2[2] =  0.0;       nx2[2] = 0.0; z2[2] = -1.0; nz2[2] = -1.0;
    makespinobject(cube,0,4,3,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}

void makepyramid(OBJECT *pyramid)
{
    x1[0] = 0.0;      nx1[0] = 2.0/sqrt(5.0);z1[0] =  1.0;nz1[0] =1.0/sqrt(5.0);
    x2[0] = sqrt(2.0);nx2[0] = 2.0/sqrt(5.0);z2[0] = -1.0;nz2[0] =1.0/sqrt(5.0);
    x1[1] = sqrt(2.0);nx1[1] = 0.0;          z1[1] = -1.0;nz1[1] = -1.0;
    x2[1] = 0.0;      nx2[1] = 0.0;          z2[1] = -1.0;nz2[1] = -1.0;
    makespinobject(pyramid,0,4,2,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}


void makeglass(OBJECT *glass, int n)
{
    x1[ 0] = 0.000; z1[ 0] = 0.575; x1[ 1] = 0.050; z1[ 1] = 0.575;
    x1[ 2] = 0.100; z1[ 2] = 0.600; x1[ 3] = 0.200; z1[ 3] = 0.700;
    x1[ 4] = 0.250; z1[ 4] = 0.800; x1[ 5] = 0.250; z1[ 5] = 1.000;
    x1[ 6] = 0.275; z1[ 6] = 1.000; x1[ 7] = 0.275; z1[ 7] = 0.800;
    x1[ 8] = 0.250; z1[ 8] = 0.700; x1[ 9] = 0.200; z1[ 9] = 0.625;
    x1[10] = 0.100; z1[10] = 0.550; x1[11] = 0.050; z1[11] = 0.500;
    x1[12] = 0.025; z1[12] = 0.450; x1[13] = 0.025; z1[13] = 0.100;
    x1[14] = 0.050; z1[14] = 0.050; x1[15] = 0.250; z1[15] = 0.025;
    x1[16] = 0.250; z1[16] = 0.000; x1[17] = 0.225; z1[17] = 0.000;
    x1[18] = 0.000; z1[18] = 0.000;
    makesmoothspinobject(glass,n,19,x1,z1);
}

void makeobjects()
/* Creates all the objects. */
{
    makecube(&cube);
    makesphere(&sphere,12);
    makecylinder(&cylinder,12);
    makecone(&cone,12);
    makeglass(&glass,18);
    makepyramid(&pyramid);
}
    
void drawobject(OBJECT obj, int wire, int flat)
/* Draws a defined object. wire indicates whether wireframe. 
   flat indicates whether flat shaded. */
{
    int i,j;
    PATCH *p;
    p = obj.pat;
    if (flat) {shademodel(FLAT);} else {shademodel(GOURAUD);};
    for (i=0; i < obj.numb ; i++)
    {
	if (wire) {bgnclosedline();} else {bgnpolygon();};
	for (j=0; j < p->numb; j++)
	    { n3f(p->n[j].c); v3f(p->v[j].c);};
	if (wire) {endclosedline();} else {endpolygon();};
	p++;
    }
}

/******************************* THE DRAWING ********************************/

#define MAXOBJ 20

int frames,back,wire,flat;      /* toggles to indicate whay of display */
int ob[MAXOBJ];                 /* indicates which objects to show */
int lcount;                     /* last frame cout */
float time_passed;              /* the "time" so far */
float speed;			/* the rotation speed */
long win1;                      /* the window identifier */

void drw(OBJECT obj, float r1, float r2, float r3,
	float t1, float t2, float t3, float s1, float s2, float s3, int mat)
/* draws a particular object. */
{
    pushmatrix();
    rot(time_passed*r1,'X'); rot(time_passed*r2,'Y'); rot(time_passed*r3,'Z');
    translate(t1,t2,t3); scale(s1,s2,s3);
    lmbind(MATERIAL,mat); drawobject(obj,wire,flat);
    popmatrix();
}

void redraw()
/* draws all the objects in the current window */
{
    char str[100];
    backface(back);
    /* clear the screen */
    zclear();
    RGBcolor(43,169,255);
    clear();
    /* do the drawing */
    if (ob[0]) drw(      sphere,  3.0,3.1,2.0,  2.2, 1.2, 2.0,  0.8,0.8,0.8, 1);
    if (ob[1]) drw(    cylinder,  3.2,2.6,1.8, -1.9, 1.2, 1.6,  0.3,0.3,1.0, 4);
    if (ob[2]) drw(        cube,  2.2,3.3,1.4, -1.0, 1.2,-1.5,  0.6,0.6,0.6, 3);
    if (ob[3]) drw(        cone,  4.2,2.1,3.2,  1.2, 1.3, 1.0,  0.5,0.5,0.8, 2);
    if (ob[3]) drw(      sphere,  4.2,2.1,3.2,  1.2, 1.3, 0.0,  0.5,0.5,0.5,51);
    if (ob[4]) drw(      sphere,  1.2,3.3,2.4, -2.0, 1.4,-2.1,  0.8,0.8,0.2, 5);
    if (ob[5]) drw(        cube,  3.2,3.6,2.4,  1.1, 1.6, 2.5,  0.8,0.3,0.1, 6);
    if (ob[6]) drw(       glass,  2.7,1.9,3.1,  1.6, 1.3, 0.7,  2.0,2.0,2.0, 7);
    if (ob[7]) drw(     pyramid,  2.3,2.7,3.3,  1.1, 2.3, 1.7,  0.6,0.6,0.5, 8);
    /* show frame count if required */
    if (frames && winget() == win1)
    {
    	cpack(0xffffff);
	cmov(-2.1,-2.1,4.0);
	sprintf(str,"%d frames per second",lcount);
	charstr(str);
    };
    /* ready */
    swapbuffers();
}

/* The identity matrix. */
float idmat[4][4] = {{1.0,0.0,0.0,0.0},
	        {0.0,1.0,0.0,0.0},
		{0.0,0.0,1.0,0.0},
		{0.0,0.0,0.0,1.0}};

/* The different materials. */
float mater1[] = 
    {SPECULAR,0.8,0.0,0.0,DIFFUSE,0.4,0.0,0.0,SHININESS,40.0,LMNULL};
float mater51[] = 
    {SPECULAR,0.2,0.0,0.1,DIFFUSE,0.8,0.0,0.3,SHININESS,10.0,LMNULL};
float mater2[] =
    {SPECULAR,1.0,0.4,0.0,DIFFUSE,1.0,0.4,0.0,SHININESS,80.0,LMNULL};
float mater3[] =
    {SPECULAR,0.0,0.0,0.6,DIFFUSE,0.0,0.0,0.8,SHININESS,60.0,LMNULL};
float mater4[] =
    {SPECULAR,0.0,1.0,0.0,DIFFUSE,0.0,0.6,0.0,SHININESS,120.0,LMNULL};
float mater5[] =
    {SPECULAR,1.0,1.0,0.0,DIFFUSE,0.6,0.6,0.0,SHININESS,100.0,LMNULL};
float mater6[] =
    {SPECULAR,1.0,0.0,1.0,DIFFUSE,0.6,0.0,0.6,SHININESS,120.0,LMNULL};
float mater7[] =
    {SPECULAR,0.9,0.9,0.9,DIFFUSE,0.6,0.6,0.6,SHININESS,120.0,LMNULL};
float mater8[] =
    {SPECULAR,0.4,0.7,0.4,DIFFUSE,0.5,1.0,0.5,SHININESS,50.0,LMNULL};

/* The lightsources. */
float light1[] = {LCOLOR,1.0,1.0,1.0,POSITION,10.0,10.0,5.0,0.0,LMNULL};
float light2[] = {LCOLOR,1.0,1.0,1.0,POSITION,-10.0,10.0,5.0,0.0,LMNULL};

/* The lightmodel. */
float model[] = {AMBIENT,0.4,0.4,0.4,LMNULL};

void init()
/* Initializes all settings for a window. */
{
    zbuffer(TRUE);
    doublebuffer();
    RGBmode();
    gconfig();
    mmode(MVIEWING);
    perspective(400,1.0,1.0,34.0);
    loadmatrix(idmat);
    /* set up the devices */
    fl_qdevice(ESCKEY);
    fl_qdevice(WINFREEZE);
    fl_qdevice(WINTHAW);
    fl_qdevice(REDRAWICONIC);
    /* define materials and lights */
    lmdef(DEFMATERIAL, 1, 11, mater1);
    lmdef(DEFMATERIAL,51, 11, mater51);
    lmdef(DEFMATERIAL, 2, 11, mater2);
    lmdef(DEFMATERIAL, 3, 11, mater3);
    lmdef(DEFMATERIAL, 4, 11, mater4);
    lmdef(DEFMATERIAL, 5, 11, mater5);
    lmdef(DEFMATERIAL, 6, 11, mater6);
    lmdef(DEFMATERIAL, 7, 11, mater7);
    lmdef(DEFMATERIAL, 8, 11, mater8);
    lmdef(DEFLIGHT, 1, 10, light1);
    lmdef(DEFLIGHT, 2, 10, light2);
    lmdef(DEFLMODEL, 1, 5, model);
    lmbind(LIGHT0,1);
    lmbind(LIGHT1,2);
    lmbind(LMODEL,1);
}

/***************************** THE FORM *********************************/

void object_cb(FL_OBJECT *obj,long arg)
{
  if (fl_get_browser(obj)>0) ob[fl_get_browser(obj)-1] = 1;
  if (fl_get_browser(obj)<0) ob[-fl_get_browser(obj)-1] = 0;
}

void wire_cb(FL_OBJECT *obj, long arg)   { wire = ! wire; }

void flat_cb(FL_OBJECT *obj, long arg)   { flat = ! flat; }

void back_cb(FL_OBJECT *obj, long arg)   { back = ! back; }

void frames_cb(FL_OBJECT *obj, long arg) { frames = ! frames; }

void theend_cb(FL_OBJECT *obj, long arg)  { exit(0); }

void speed_cb(FL_OBJECT *obj, long arg) { speed = fl_get_slider(obj); }

void makeform()
{
  create_the_forms();
  fl_add_browser_line(objects_obj,"Sphere");
  fl_select_browser_line(objects_obj,1);
  fl_add_browser_line(objects_obj,"Cylinder");
  fl_select_browser_line(objects_obj,2);
  fl_add_browser_line(objects_obj,"Cube");
  fl_select_browser_line(objects_obj,3);
  fl_add_browser_line(objects_obj,"Ice Cream");
  fl_add_browser_line(objects_obj,"Disk");
  fl_add_browser_line(objects_obj,"Diamond");
  fl_add_browser_line(objects_obj,"Glass");
  fl_add_browser_line(objects_obj,"Pyramid");
  fl_set_button(wire_obj,1);
  fl_set_button(flat_obj,1);
  fl_set_button(back_obj,1);
  fl_set_button(frames_obj,1);
  fl_set_slider_bounds(speed_obj,0.0,20.0);
  fl_set_slider_value(speed_obj,5.0);
}

main()
{
    long dev;
    short val;
    struct timeval tp;
    struct timezone tzp;
    long passed,lastsec,lasttime;
    int count = 0;
    int i;
    FL_OBJECT *retobj;
    /* start up the stuff */
    makeform();
    fl_show_form(form,FL_PLACE_SIZE,FALSE,NULL);
    makeobjects();
    prefposition(100,500,500,900);
    win1 = winopen("Objects");
    winconstraints(); iconsize(85,66); keepaspect(1,1); winconstraints();
    init();
    lookat(0.0,0.0,10.0,0.0,0.0,0.0,0);
    for (i=0; i<MAXOBJ; i++) ob[i] = 0;
    ob[0] = ob[1] = ob[2] = 1;
    frames = 1; back = 1; wire = 0; flat = 0; time_passed= 0.0;
    speed = 5.0;
    /* ready to go */
    gettimeofday(&tp,&tzp);
    lastsec = tp.tv_sec;
    lasttime = tp.tv_usec;
    while (1)
    {
	/* draw the stuff */
	gettimeofday(&tp,&tzp);
	passed = 1000000 * (tp.tv_sec - lastsec) + (tp.tv_usec - lasttime);
	time_passed += speed * passed/250000.0;
        if (lastsec != tp.tv_sec) {lcount = count; count = 0;};
        count++;
        lastsec = tp.tv_sec;
        lasttime = tp.tv_usec;
	winset(win1); redraw();
	/* listen at the devices */
	retobj = fl_check_forms();
	if (retobj == FL_EVENT)
	{
	    dev = fl_qread(&val);
	    if (dev == ESCKEY) { exit(0); }
	    else if (dev == REDRAW)
	        { winset(val); reshapeviewport();}
	    else if (dev == REDRAWICONIC)
	        { winset(val); reshapeviewport();}
	    else if (dev == WINTHAW)
	        { winset(val); reshapeviewport();}
        }
    }
}
